@@ -16,7 +16,9 @@ struct _IBusM17NEngine {
     MInputContext *context;
     IBusLookupTable *table;
     IBusProperty    *status_prop;
+    IBusProperty    *setup_prop;
     IBusPropList    *prop_list;
+    gchar *config_section;
 };
 
 struct _IBusM17NEngineClass {
@@ -70,10 +72,28 @@ static void ibus_m17n_engine_commit_string
                                              const gchar            *string);
 static void ibus_m17n_engine_callback       (MInputContext          *context,
                                              MSymbol                 command);
+static void ibus_m17n_engine_update_preedit (IBusM17NEngine *m17n);
+static void ibus_m17n_engine_update_lookup_table
+                                            (IBusM17NEngine *m17n);
 
 static IBusEngineClass *parent_class = NULL;
 static GHashTable      *im_table = NULL;
 
+static IBusConfig      *config = NULL;
+static guint            preedit_foreground = INVALID_COLOR;
+static guint            preedit_background = INVALID_COLOR;
+static gint             preedit_underline = IBUS_ATTR_UNDERLINE_NONE;
+static gint             lookup_table_orientation = IBUS_ORIENTATION_SYSTEM;
+
+void
+ibus_m17n_init (IBusBus *bus)
+{
+    config = ibus_bus_get_config (bus);
+    if (config)
+        g_object_ref_sink (config);
+    ibus_m17n_init_common ();
+}
+
 GType
 ibus_m17n_engine_get_type (void)
 {
@@ -131,9 +151,49 @@ ibus_m17n_engine_class_init (IBusM17NEngineClass *klass)
     engine_class->property_activate = ibus_m17n_engine_property_activate;
 }
 
+static void
+ibus_config_value_changed (IBusConfig   *config,
+                           const gchar  *section,
+                           const gchar  *name,
+                           GValue       *value,
+                           gpointer      user_data)
+{
+    IBusM17NEngine *m17n = (IBusM17NEngine *) user_data;
+
+    if (g_strcmp0 (section, m17n->config_section) == 0) {
+        if (g_strcmp0 (name, "preedit_foreground") == 0) {
+            const gchar *hex = g_value_get_string (value);
+            guint color;
+            color = ibus_m17n_parse_color (hex);
+            if (color != INVALID_COLOR) {
+                preedit_foreground = color;
+                ibus_m17n_engine_update_preedit (m17n);
+            }
+        } else if (g_strcmp0 (name, "preedit_background") == 0) {
+            const gchar *hex = g_value_get_string (value);
+            guint color;
+            color = ibus_m17n_parse_color (hex);
+            if (color != INVALID_COLOR) {
+                preedit_background = color;
+                ibus_m17n_engine_update_preedit (m17n);
+            }
+        } else if (g_strcmp0 (name, "preedit_underline") == 0) {
+            preedit_underline = g_value_get_int (value);
+            ibus_m17n_engine_update_preedit (m17n);
+        } else if (g_strcmp0 (name, "lookup_table_orientation") == 0) {
+            lookup_table_orientation = g_value_get_int (value);
+            ibus_m17n_engine_update_lookup_table (m17n);
+        }
+    }
+}
+
 static void
 ibus_m17n_engine_init (IBusM17NEngine *m17n)
 {
+    IBusText* label;
+    IBusText* tooltip;
+    const gchar *engine_name;
+
     m17n->status_prop = ibus_property_new ("status",
                                            PROP_TYPE_NORMAL,
                                            NULL,
@@ -145,9 +205,23 @@ ibus_m17n_engine_init (IBusM17NEngine *m17n)
                                            NULL);
     g_object_ref_sink (m17n->status_prop);
 
+    label = ibus_text_new_from_string ("Setup");
+    tooltip = ibus_text_new_from_string ("Configure M17N engine");
+    m17n->setup_prop = ibus_property_new ("setup",
+                                          PROP_TYPE_NORMAL,
+                                          label,
+                                          "gtk-preferences",
+                                          tooltip,
+                                          TRUE,
+                                          TRUE,
+                                          PROP_STATE_UNCHECKED,
+                                          NULL);
+    g_object_ref_sink (m17n->setup_prop);
+
     m17n->prop_list = ibus_prop_list_new ();
     g_object_ref_sink (m17n->prop_list);
     ibus_prop_list_append (m17n->prop_list,  m17n->status_prop);
+    ibus_prop_list_append (m17n->prop_list, m17n->setup_prop);
 
     m17n->table = ibus_lookup_table_new (9, 0, TRUE, TRUE);
     g_object_ref_sink (m17n->table);
@@ -162,6 +236,11 @@ ibus_m17n_engine_constructor (GType                   type,
     IBusM17NEngine *m17n;
     MInputMethod *im;
     const gchar *engine_name;
+    gchar *lang;
+    gchar *name;
+    gchar **strv;
+    GValue value = { 0 };
+    gboolean preedit_highlight;
 
     m17n = (IBusM17NEngine *) G_OBJECT_CLASS (parent_class)->constructor (type,
                                                        n_construct_params,
@@ -170,6 +249,14 @@ ibus_m17n_engine_constructor (GType                   type,
     engine_name = ibus_engine_get_name ((IBusEngine *) m17n);
     g_assert (engine_name);
 
+    strv = g_strsplit (engine_name, ":", 3);
+
+    g_assert (g_strv_length (strv) == 3);
+    g_assert (g_strcmp0 (strv[0], "m17n") == 0);
+
+    lang = strv[1];
+    name = strv[2];
+
     if (im_table == NULL) {
         im_table = g_hash_table_new_full (g_str_hash,
                                           g_str_equal,
@@ -179,18 +266,6 @@ ibus_m17n_engine_constructor (GType                   type,
 
     im = (MInputMethod *) g_hash_table_lookup (im_table, engine_name);
     if (im == NULL) {
-        gchar *lang;
-        gchar *name;
-        gchar **strv;
-
-        strv = g_strsplit (engine_name, ":", 3);
-
-        g_assert (g_strv_length (strv) == 3);
-        g_assert (g_strcmp0 (strv[0], "m17n") == 0);
-
-        lang = strv[1];
-        name = strv[2];
-
         im = minput_open_im (msymbol (lang), msymbol (name), NULL);
         if (im != NULL) {
             mplist_put (im->driver.callback_list, Minput_preedit_start, ibus_m17n_engine_callback);
@@ -210,12 +285,11 @@ ibus_m17n_engine_constructor (GType                   type,
 
             g_hash_table_insert (im_table, g_strdup (engine_name), im);
         }
-
-        g_strfreev (strv);
     }
 
     if (im == NULL) {
         g_warning ("Can not find m17n keymap %s", engine_name);
+        g_strfreev (strv);
         g_object_unref (m17n);
         return NULL;
     }
@@ -223,6 +297,56 @@ ibus_m17n_engine_constructor (GType                   type,
     m17n->context = minput_create_ic (im, NULL);
     mplist_add (m17n->context->plist, msymbol ("IBusEngine"), m17n);
 
+    m17n->config_section = g_strdup_printf ("engine/M17N/%s/%s",
+                                            lang, name);
+    g_strfreev (strv);
+
+    preedit_highlight = ibus_m17n_preedit_highlight (engine_name);
+    if (ibus_config_get_value (config,
+                               m17n->config_section,
+                               "preedit_foreground",
+                               &value)) {
+        const gchar *hex = g_value_get_string (&value);
+        guint color = ibus_m17n_parse_color (hex);
+        if (color != (guint)-1)
+            preedit_foreground = color;
+        g_value_unset (&value);
+    } else if (preedit_highlight)
+        preedit_foreground = PREEDIT_FOREGROUND;
+
+    if (ibus_config_get_value (config,
+                               m17n->config_section,
+                               "preedit_background",
+                               &value)) {
+        const gchar *hex = g_value_get_string (&value);
+        guint color = ibus_m17n_parse_color (hex);
+        if (color != (guint)-1)
+            preedit_background = color;
+        g_value_unset (&value);
+    } else if (preedit_highlight)
+        preedit_background = PREEDIT_BACKGROUND;
+
+    if (ibus_config_get_value (config,
+                               m17n->config_section,
+                               "preedit_underline",
+                               &value)) {
+        preedit_underline = g_value_get_int (&value);
+        g_value_unset (&value);
+    } else
+        preedit_underline = IBUS_ATTR_UNDERLINE_NONE;
+
+    if (ibus_config_get_value (config,
+                               m17n->config_section,
+                               "lookup_table_orientation",
+                               &value)) {
+        lookup_table_orientation = g_value_get_int (&value);
+        g_value_unset (&value);
+    } else
+        lookup_table_orientation = IBUS_ORIENTATION_SYSTEM;
+
+    g_signal_connect (config, "value-changed",
+                      G_CALLBACK(ibus_config_value_changed), m17n);
+
     return (GObject *) m17n;
 }
 
@@ -240,6 +364,11 @@ ibus_m17n_engine_destroy (IBusM17NEngine *m17n)
         m17n->status_prop = NULL;
     }
 
+    if (m17n->setup_prop) {
+        g_object_unref (m17n->setup_prop);
+        m17n->setup_prop = NULL;
+    }
+
     if (m17n->table) {
         g_object_unref (m17n->table);
         m17n->table = NULL;
@@ -250,9 +379,38 @@ ibus_m17n_engine_destroy (IBusM17NEngine *m17n)
         m17n->context = NULL;
     }
 
+    if (m17n->config_section) {
+        g_free (m17n->config_section);
+        m17n->config_section = NULL;
+    }
+
     IBUS_OBJECT_CLASS (parent_class)->destroy ((IBusObject *)m17n);
 }
 
+static void
+ibus_m17n_engine_update_preedit (IBusM17NEngine *m17n)
+{
+    IBusText *text;
+    gchar *buf;
+
+    buf = ibus_m17n_mtext_to_utf8 (m17n->context->preedit);
+    if (buf) {
+        text = ibus_text_new_from_static_string (buf);
+        if (preedit_foreground != INVALID_COLOR)
+            ibus_text_append_attribute (text, IBUS_ATTR_TYPE_FOREGROUND,
+                                        preedit_foreground, 0, -1);
+        if (preedit_background != INVALID_COLOR)
+            ibus_text_append_attribute (text, IBUS_ATTR_TYPE_BACKGROUND,
+                                        preedit_background, 0, -1);
+        ibus_text_append_attribute (text, IBUS_ATTR_TYPE_UNDERLINE,
+                                    preedit_underline, 0, -1);
+        ibus_engine_update_preedit_text ((IBusEngine *) m17n,
+                                         text,
+                                         m17n->context->cursor_pos,
+                                         mtext_len (m17n->context->preedit) > 0);
+    }
+}
+
 static void
 ibus_m17n_engine_commit_string (IBusM17NEngine *m17n,
                                 const gchar    *string)
@@ -260,24 +418,48 @@ ibus_m17n_engine_commit_string (IBusM17NEngine *m17n,
     IBusText *text;
     text = ibus_text_new_from_static_string (string);
     ibus_engine_commit_text ((IBusEngine *)m17n, text);
+    ibus_m17n_engine_update_preedit (m17n);
 }
 
+/* Note on AltGr handling: While currently we expect AltGr == mod5, it
+   would be better to not expect the modifier always be assigned
+   to particular modX.  However, it needs some code like:
+
+   KeyCode altgr = XKeysymToKeycode (display, XK_ISO_Level3_Shift);
+   XModifierKeymap *mods = XGetModifierMapping (display);
+   for (i = 3; i < 8; i++)
+     for (j = 0; j < mods->max_keypermod; j++) {
+       KeyCode code = mods->modifiermap[i * mods->max_keypermod + j];
+       if (code == altgr)
+         ...
+     }
+                
+   Since IBus engines are supposed to be cross-platform, the code
+   should go into IBus core, instead of ibus-m17n. */
 MSymbol
-ibus_m17n_key_event_to_symbol (guint keyval,
+ibus_m17n_key_event_to_symbol (guint keycode,
+                               guint keyval,
                                guint modifiers)
 {
     GString *keysym;
     MSymbol mkeysym = Mnil;
     guint mask = 0;
+    IBusKeymap *keymap;
+    guint base_keyval;
 
     if (keyval >= IBUS_Shift_L && keyval <= IBUS_Hyper_R) {
         return Mnil;
     }
 
+    keymap = ibus_keymap_get ("us");
+    base_keyval = ibus_keymap_lookup_keysym (keymap, keycode, 0);
+    g_object_unref (keymap);
+
     keysym = g_string_new ("");
 
-    if (keyval >= IBUS_space && keyval <= IBUS_asciitilde) {
-        gint c = keyval;
+    if (base_keyval >= IBUS_space && base_keyval <= IBUS_asciitilde) {
+        gint c = (modifiers & IBUS_MOD5_MASK) ? base_keyval : keyval;
+
         if (keyval == IBUS_space && modifiers & IBUS_SHIFT_MASK)
             mask |= IBUS_SHIFT_MASK;
 
@@ -299,6 +481,7 @@ ibus_m17n_key_event_to_symbol (guint keyval,
     }
 
     mask |= modifiers & (IBUS_MOD1_MASK |
+                         IBUS_MOD5_MASK |
                          IBUS_META_MASK |
                          IBUS_SUPER_MASK |
                          IBUS_HYPER_MASK);
@@ -310,6 +493,9 @@ ibus_m17n_key_event_to_symbol (guint keyval,
     if (mask & IBUS_SUPER_MASK) {
         g_string_prepend (keysym, "s-");
     }
+    if (mask & IBUS_MOD5_MASK) {
+        g_string_prepend (keysym, "G-");
+    }
     if (mask & IBUS_MOD1_MASK) {
         g_string_prepend (keysym, "A-");
     }
@@ -372,7 +558,7 @@ ibus_m17n_engine_process_key_event (IBusEngine     *engine,
 
     if (modifiers & IBUS_RELEASE_MASK)
         return FALSE;
-    MSymbol m17n_key = ibus_m17n_key_event_to_symbol (keyval, modifiers);
+    MSymbol m17n_key = ibus_m17n_key_event_to_symbol (keycode, keyval, modifiers);
 
     if (m17n_key == Mnil)
         return FALSE;
@@ -471,6 +657,19 @@ ibus_m17n_engine_property_activate (IBusEngine  *engine,
                                     const gchar *prop_name,
                                     guint        prop_state)
 {
+    IBusM17NEngine *m17n = (IBusM17NEngine *) engine;
+
+    if (g_strcmp0 (prop_name, "setup") == 0) {
+        const gchar *engine_name;
+        gchar *setup;
+
+        engine_name = ibus_engine_get_name ((IBusEngine *) m17n);
+        g_assert (engine_name);
+        setup = g_strdup_printf ("%s/ibus-setup-m17n --name %s",
+                                 LIBEXECDIR, engine_name);
+        g_spawn_command_line_async (setup, NULL);
+        g_free (setup);
+    }
     parent_class->property_activate (engine, prop_name, prop_state);
 }
 
@@ -503,14 +702,15 @@ ibus_m17n_engine_update_lookup_table (IBusM17NEngine *m17n)
 
         if (mplist_key (group) == Mtext) {
             MText *mt;
-            gunichar *buf, *p;
+            gunichar *buf;
+            glong nchars, i;
 
             mt = (MText *) mplist_value (group);
             ibus_lookup_table_set_page_size (m17n->table, mtext_len (mt));
 
-            buf = ibus_m17n_mtext_to_ucs4 (mt);
-            for (p = buf + 1; *p != 0; p++) {
-                ibus_lookup_table_append_candidate (m17n->table, ibus_text_new_from_unichar (*p));
+            buf = ibus_m17n_mtext_to_ucs4 (mt, &nchars);
+            for (i = 0; i < nchars; i++) {
+                ibus_lookup_table_append_candidate (m17n->table, ibus_text_new_from_unichar (buf[i]));
             }
             g_free (buf);
         }
@@ -534,6 +734,8 @@ ibus_m17n_engine_update_lookup_table (IBusM17NEngine *m17n)
         }
 
         ibus_lookup_table_set_cursor_pos (m17n->table, m17n->context->candidate_index - i);
+        ibus_lookup_table_set_orientation (m17n->table, lookup_table_orientation);
+
         text = ibus_text_new_from_printf ("( %d / %d )", page, mplist_length (m17n->context->candidate_list));
 
         ibus_engine_update_lookup_table ((IBusEngine *)m17n, m17n->table, TRUE);
@@ -565,19 +767,7 @@ ibus_m17n_engine_callback (MInputContext *context,
         ibus_engine_hide_preedit_text ((IBusEngine *)m17n);
     }
     else if (command == Minput_preedit_draw) {
-        IBusText *text;
-        gchar *buf;
-
-        buf = ibus_m17n_mtext_to_utf8 (m17n->context->preedit);
-        if (buf) {
-            text = ibus_text_new_from_static_string (buf);
-            ibus_text_append_attribute (text, IBUS_ATTR_TYPE_FOREGROUND, 0x00ffffff, 0, -1);
-            ibus_text_append_attribute (text, IBUS_ATTR_TYPE_BACKGROUND, 0x00000000, 0, -1);
-            ibus_engine_update_preedit_text ((IBusEngine *) m17n,
-                                             text,
-                                             m17n->context->cursor_pos,
-                                             mtext_len (m17n->context->preedit) > 0);
-        }
+        ibus_m17n_engine_update_preedit (m17n);
     }
     else if (command == Minput_preedit_done) {
         ibus_engine_hide_preedit_text ((IBusEngine *)m17n);
